package arzumify.networked.impl;

import arzumify.presence.maths.Vec2i;
import arzumify.presence.maths.Vec3i;
import arzumify.presence.presences.IntegerPresence;
import arzumify.presence.presences.Presence;

import java.util.*;

public class ModifiablePresence extends IntegerPresence implements Presence<IntegerPresence> {
    public ModifiablePresence(short range, Vec3i... basePositions) {
        super(range, basePositions);
        this.basePositions.addAll(List.of(basePositions));
        this.range = range;
        this.baseChunkPresence = getChunkPresence(this.basePositions);
        this.collisionChunkPresence = getChunkPresence(ComputeCollisionShape(range, this.basePositions));
    }

    public final HashSet<Vec3i> basePositions = new HashSet<>();
    public HashSet<Vec2i> baseChunkPresence;
    public HashSet<Vec2i> collisionChunkPresence;
    private final short range;

    public void addNewBasePosition(Set<Vec3i> positions) {
        basePositions.addAll(positions);
        baseChunkPresence.addAll(getChunkPresence(positions));
        collisionChunkPresence.addAll(getChunkPresence(ComputeCollisionShape(range, positions)));
    }

    // Calculates the chunk presence for the given positions
    public HashSet<Vec2i> getChunkPresence(Iterable<Vec3i> positions) {
        var chunkPresence = new HashSet<Vec2i>();
        for (Vec3i basePosition : positions) {
            chunkPresence.add(new Vec2i(basePosition.x() >> 4, basePosition.z() >> 4));
        }
        return chunkPresence;
    }

    @Override
    public boolean Matches(Presence<IntegerPresence> presence) {
        return presence instanceof ModifiablePresence && matches((ModifiablePresence) presence);
    }

    private boolean matches(ModifiablePresence other) {
        // Take the other's base position and calculate the manhattan distance to each of this presence's positions
        for (Vec3i otherBasePosition : other.basePositions) {
            for (Vec3i presence : this.basePositions) {
                if (ManhattanDistance(otherBasePosition, presence) <= this.range) {
                    return true;
                } else {
                    // Print the manhattan distance
                    System.out.println("Manhattan distance between " + otherBasePosition + " and " + presence + " is " + ManhattanDistance(otherBasePosition, presence));
                }
            }
        }
        return false;
    }

    private static final Map<Short, HashSet<Vec3i>> cachedCollisionShapes = new HashMap<>();

    private static HashSet<Vec3i> ComputeCollisionShape(short range, Set<Vec3i> basePositions) {
        if (!cachedCollisionShapes.containsKey(range)) {
            var possiblePresences = ComputePossiblePresences(range);
            cachedCollisionShapes.put(range, possiblePresences);
        }
        var collisionShape = new HashSet<Vec3i>();
        for (Vec3i basePosition : basePositions) {
            for (Vec3i possiblePresence : cachedCollisionShapes.get(range)) {
                collisionShape.add(basePosition.add(possiblePresence));
            }
        }
        return collisionShape;
    }

    public static HashSet<Vec3i> ComputePossiblePresences(short range) {
        var presences = new HashSet<Vec3i>();
        for (int dx = -range; dx <= range; dx++) {
            for (int dy = -range; dy <= range; dy++) {
                int dzMax = range - (Math.abs(dx) + Math.abs(dy));
                if (dzMax >= 0) {
                    for (int dz = -dzMax; dz <= dzMax; dz++) {
                        presences.add(new Vec3i(dx, dy, dz));
                    }
                }
            }
        }
        return presences;
    }

    @Override
    public HashSet<Vec2i> GetChunks() {
        return baseChunkPresence;
    }

    @Override
    public HashSet<Vec2i> GetCollisionChunks() {
        return collisionChunkPresence;
    }
}
